#include <Wire.h>
#include <MadgwickAHRS.h>

#define RFID_TAG ("7F0007FB43C0")

#define ADXL345_ADDRESS (0xA6 >> 1)
#define ADXL345_REGISTER_XLSB (0x32)
#define ADXL_REGISTER_PWRCTL (0x2D)
#define ADXL_DATA_FORMAT (0x31)
#define ADXL_OFSX (0x1E)
#define ADXL_OFSY (0x1F)
#define ADXL_OFSZ (0x20)
#define ADXL_INT_ENABLE (0x2E)
#define ADXL_THRESH_TAP (0x1D)
#define ADXL_THRESH_ACT (0x24)
#define ADXL_THRESH_INACT (0x25)
#define ADXL_TIME_INACT (0x26)
#define ADXL_DUR (0x21)
#define ADXL_FIFO_CTL (0x38)
#define ADXL_PWRCTL_MEASURE (1 << 3)

#define ITG3200_ADDRESS (0xD0 >> 1)
#define ITG3200_REGISTER_XMSB (0x1D)
#define ITG3200_REGISTER_DLPF_FS (0x16)
#define ITG3200_FULLSCALE (0x03 << 3)
#define ITG3200_42HZ (0x03)

#define HMC5843_ADDRESS (0x3C >> 1)
#define HMC5843_REGISTER_XMSB (0x03)
#define HMC5843_REGISTER_MEASMODE (0x02)
#define HMC5843_MEASMODE_CONT (0x00)

#define sampleFreq   25.0f      // sample frequency in Hz

int accelerometer_data[3];
int gyro_data[3];
int magnetometer_data[3];

const int down_btn = 17;
const int left_btn = 4;
const int right_btn = 10;
const int up_btn = 9;
const int a_btn = 8;
const int b_btn = 7;
const int start_btn = 6;
const int select_btn = 5;

int btn1 = 0;
int btn2 = 0;
int btn3 = 0;
int btn4 = 0;
int btn5 = 0;
int btn6 = 0;
int btn7 = 0;
int btn8 = 0;

int send_flag = 0;

Madgwick filter; // initialise Madgwick object
float yaw, pitch, roll;
float factor = 800; // variable by which to divide gyroscope values, used to control sensitivity

char c;

void setup() {
  send_flag = 0;
  
  pinMode(down_btn, INPUT);
  digitalWrite(down_btn, HIGH);
  pinMode(left_btn, INPUT);
  digitalWrite(left_btn, HIGH);
  pinMode(right_btn, INPUT);
  digitalWrite(right_btn, HIGH);
  pinMode(up_btn, INPUT);
  digitalWrite(up_btn, HIGH);
  pinMode(a_btn, INPUT);
  digitalWrite(a_btn, HIGH);
  pinMode(b_btn, INPUT);
  digitalWrite(b_btn, HIGH);
  pinMode(start_btn, INPUT);
  digitalWrite(start_btn, HIGH);
  pinMode(select_btn, INPUT);
  digitalWrite(select_btn, HIGH);
  
  Wire.begin();
  Serial.begin(9600);
  Serial1.begin(9600);

  for(int i = 0; i < 3; ++i) {
    accelerometer_data[i] = magnetometer_data[i] = gyro_data[i] = 0;
  }
  
  init_adxl345();
  init_hmc5843();
  init_itg3200();

  filter.begin(sampleFreq);
}

void loop() {
  btn1 = digitalRead(down_btn);
  btn2 = digitalRead(left_btn);
  btn3 = digitalRead(right_btn);
  btn4 = digitalRead(up_btn);
  btn5 = digitalRead(a_btn);
  btn6 = digitalRead(b_btn);
  btn7 = digitalRead(start_btn);
  btn8 = digitalRead(select_btn);  

  if (send_flag == 0) //If send data flag is false check if any buttons are pressed to show the controller isnt on by accident or charging)
  {
    if (btn1 == 0) //Down Button is pressed
    {
      send_flag = 1;
    }
    if (btn2 == 0) //Left Button is pressed
    {
      send_flag = 1;
    }
    if (btn3 == 0) //Right Button is pressed
    {
      send_flag = 1;
    }
    if (btn4 == 0) //Up Button is pressed
    {
      send_flag = 1;
    }
    if (btn5 == 0) //A Button is pressed
    {
      send_flag = 1;
    }
    if (btn6 == 0) //B Button is pressed
    {
      send_flag = 1;
    }
    if (btn7 == 0) //Start Button is pressed
    {
      send_flag = 1;
    }
    if (btn8 == 0) //Select Button is pressed
    {
      send_flag = 1;
    }
  }

  if (send_flag == 1) //Any Button has been pressed (not on by accident or charging) can start sending data
  {
    read_adxl345();
    read_hmc5843();
    read_itg3200();

    String output = "";

    // filter.updateIMU(gyro.value.x / factor, gyro.value.y / factor, gyro.value.z / factor, accel.value.x, accel.value.y, accel.value.z);
    filter.update(gyro_data[0] / factor, gyro_data[1] / factor, gyro_data[2] / factor, accelerometer_data[0] / factor, accelerometer_data[1] / factor, accelerometer_data[2] / factor, magnetometer_data[0] / factor, magnetometer_data[1] / factor, magnetometer_data[2] / factor);

    yaw = filter.getYaw();
    roll = filter.getRoll();
    pitch = filter.getPitch();

    output += "%";
    output += RFID_TAG;
    output += ",";
    output += btn1;
    output += ",";
    output += btn2;
    output += ",";
    output += btn3;
    output += ",";
    output += btn4;
    output += ",";
    output += btn5;
    output += ",";
    output += btn6;
    output += ",";
    output += btn7;
    output += ",";
    output += btn8;
    output += ",";
    output += accelerometer_data[0];
    output += ",";
    output += accelerometer_data[1];
    output += ",";
    output += accelerometer_data[2];
    output += ",";
    output += magnetometer_data[0];
    output += ",";
    output += magnetometer_data[1];
    output += ",";
    output += magnetometer_data[2];
    output += ",";
    output += gyro_data[0];
    output += ",";
    output += gyro_data[1];
    output += ",";
    output += gyro_data[2];
    output += ",";
    output += yaw;
    output += ",";
    output += roll;
    output += ",";
    output += pitch;
    output += ";";

    Serial1.print(output);
  }

  delay(100);
}

void i2c_write(int address, byte reg, byte data) {
  Wire.beginTransmission(address);
  Wire.write(reg);
  Wire.write(data);
  Wire.endTransmission();
}

void i2c_read(int address, byte reg, int count, byte* data) {
 int i = 0;

 Wire.beginTransmission(address);
 Wire.write(reg);
 Wire.endTransmission();
 Wire.beginTransmission(address);
 Wire.requestFrom(address,count);
 while(Wire.available()){
   c = Wire.read();
   data[i] = c;
   i++;
 }
 Wire.endTransmission();
} 

void init_adxl345() {
  byte data = 0;

  //self test
  i2c_write(ADXL345_ADDRESS, ADXL_DATA_FORMAT, 0x0b); //+-16g, 13bit mode
  i2c_write(ADXL345_ADDRESS, ADXL_REGISTER_PWRCTL, 0x08); //start measurement
  i2c_write(ADXL345_ADDRESS, ADXL_INT_ENABLE, 0x80); //enable data_ready interrupt
  delay(500); //wait 1.1ms for 100 data points and averaging
  i2c_write(ADXL345_ADDRESS, ADXL_DATA_FORMAT, 0x8B); //self-test on +-16g, 13bit mode
  delay(500); //wait 1.1ms for 100 data points and averaging
  i2c_write(ADXL345_ADDRESS, ADXL_DATA_FORMAT, 0x0b); //self-test off +-16g 13-bit mode

  i2c_write(ADXL345_ADDRESS, ADXL_DATA_FORMAT, 0x01); //Put the ADXL345 into +/- 4G range
  i2c_write(ADXL345_ADDRESS, ADXL_THRESH_TAP, 0x38); //set the tap threshold
  i2c_write(ADXL345_ADDRESS, ADXL_DUR, 0x10); //set the tap duration

  i2c_write(ADXL345_ADDRESS, ADXL_THRESH_ACT, 0x08); //set the activity threshold to 0.5g
  i2c_write(ADXL345_ADDRESS, ADXL_THRESH_INACT, 0x03); //set the inactivity threshold as 0.1875g
  i2c_write(ADXL345_ADDRESS, ADXL_TIME_INACT, 0x02); //set the inactivity time as 2s

  i2c_write(ADXL345_ADDRESS, ADXL_OFSX, 0x00); //set the x axis offset (no offset)
  i2c_write(ADXL345_ADDRESS, ADXL_OFSY, 0x00); //set the y axis offset (no offset)
  i2c_write(ADXL345_ADDRESS, ADXL_OFSZ, 0x00); //set the z axis offset (-56)

  i2c_write(ADXL345_ADDRESS, ADXL_REGISTER_PWRCTL, ADXL_PWRCTL_MEASURE);

  i2c_read(ADXL345_ADDRESS, ADXL_REGISTER_PWRCTL, 1, &data);
  Serial.println((unsigned int)data);
}

void read_adxl345() {
 byte bytes[6];
 memset(bytes,0,6);

 i2c_read(ADXL345_ADDRESS, ADXL345_REGISTER_XLSB, 6, bytes);

 for (int i=0;i<3;++i) {
 accelerometer_data[i] = (int)bytes[2*i] + (((int)bytes[2*i + 1]) << 8);
 }
}

void init_itg3200() {
  byte data = 0;

  i2c_write(ITG3200_ADDRESS, ITG3200_REGISTER_DLPF_FS, ITG3200_FULLSCALE | ITG3200_42HZ);

  i2c_read(ITG3200_ADDRESS, ITG3200_REGISTER_DLPF_FS, 1, &data);

  Serial.println((unsigned int)data);
}

void read_itg3200() {
  byte bytes[6];
  memset(bytes,0,6);

  i2c_read(ITG3200_ADDRESS, ITG3200_REGISTER_XMSB, 6, bytes);
  for (int i=0;i<3;++i) {
  gyro_data[i] = (int)bytes[2*i + 1] + (((int)bytes[2*i]) << 8);
  }
}

void init_hmc5843() {
  byte data = 0;
  
  i2c_write(HMC5843_ADDRESS, HMC5843_REGISTER_MEASMODE, HMC5843_MEASMODE_CONT);

  i2c_read(HMC5843_ADDRESS, HMC5843_REGISTER_MEASMODE, 1, &data);
  Serial.println((unsigned int)data);
}

void read_hmc5843() {
 byte bytes[6];
 memset(bytes,0,6);

 i2c_read(HMC5843_ADDRESS, HMC5843_REGISTER_XMSB, 6, bytes);

 for (int i=0;i<3;++i) {
 magnetometer_data[i] = (int)bytes[2*i + 1] + (((int)bytes[2*i]) << 8);
 }
}


